# -*- coding: utf-8 -*-
from abc import ABC
from copy import deepcopy
import numpy as np
import warnings

from ..state import BranchSupplemental, State
from .move import Move


__all__ = ["RedBlueMove"]


class RedBlueMove(Move, ABC):
    """
    An abstract red-blue ensemble move with parallelization as described in
    `Foreman-Mackey et al. (2013) <https://arxiv.org/abs/1202.3665>`_.

    This class is heavily based on the original from ``emcee``.

    Args:
        nsplits (int, optional): The number of sub-ensembles to use. Each
            sub-ensemble is updated in parallel using the other sets as the
            complementary ensemble. The default value is ``2`` and you
            probably won't need to change that.
        randomize_split (bool, optional): Randomly shuffle walkers between
            sub-ensembles. The same number of walkers will be assigned to
            each sub-ensemble on each iteration. (default: ``True``)
        live_dangerously (bool, optional): By default, an update will fail with
            a ``RuntimeError`` if the number of walkers is smaller than twice
            the dimension of the problem because the walkers would then be
            stuck on a low dimensional subspace. This can be avoided by
            switching between the stretch move and, for example, a
            Metropolis-Hastings step. If you want to do this and suppress the
            error, set ``live_dangerously = True``. Thanks goes (once again)
            to @dstndstn for this wonderful terminology. (default: ``False``)
        **kwargs (dict, optional): Kwargs for parent classes.

    """

    def __init__(
        self, nsplits=2, randomize_split=True, live_dangerously=False, **kwargs
    ):
        super(RedBlueMove, self).__init__(**kwargs)
        self.nsplits = int(nsplits)
        self.live_dangerously = live_dangerously
        self.randomize_split = randomize_split

    def setup(self, branches_coords):
        """Any setup for the proposal.

        Args:
            branches_coords (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim]. These are the curent
                coordinates for all the walkers.

        """

    @classmethod
    def get_proposal(self, sample, complement, random, gibbs_ndim=None):
        """Make a proposal

        Args:
            sample (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim].
            complement (dict): Keys are ``branch_names``. Values are lists of other
                other np.ndarray[ntemps, nwalkers - subset size, nleaves_max, ndim] from
                all other subsets. This is the compliment whose ``coords`` are
                used to form the proposal for the ``sample`` subset.
            random (object): Current random state of the sampler.
            gibbs_ndim (int or np.ndarray, optional): If Gibbs sampling, this indicates
                the true dimension. If given as an array, must have shape ``(ntemps, nwalkers)``.
                See the tutorial for more information.
                (default: ``None``)

        Returns:
            tuple: Tuple contained proposal information.
                First entry is the new coordinates as a dictionary with keys
                as ``branch_names`` and values as np.ndarray[ntemps, nwalkers, nleaves_max, ndim]
                of new coordinates. Second entry is the factors associated with the
                proposal necessary for detailed balance. This is effectively
                any term in the detailed balance fraction. +log of factors if
                in the numerator. -log of factors if in the denominator.

        """

        raise NotImplementedError("The proposal must be implemented by " "subclasses")

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            tuple: (state, accepted)
                The first return is the state of the sampler after the move.
                The second return value is the accepted count array.

        """
        # Check that the dimensions are compatible.
        ndim_total = 0
        for branch in state.branches.values():
            ntemps, nwalkers, nleaves_, ndim_ = branch.shape
            ndim_total += ndim_ * nleaves_

        if nwalkers < 2 * ndim_total and not self.live_dangerously:
            raise RuntimeError(
                "It is unadvisable to use a red-blue move "
                "with fewer walkers than twice the number of "
                "dimensions. If you would like to do this, please set live_dangerously"
                "to True."
            )

        # Run any move-specific setup.
        self.setup(state.branches)

        # Split the ensemble in half and iterate over these two halves.
        accepted = np.zeros((ntemps, nwalkers), dtype=bool)
        all_inds = np.tile(np.arange(nwalkers), (ntemps, 1))
        inds = all_inds % self.nsplits
        if self.randomize_split:
            [np.random.shuffle(x) for x in inds]

        all_branch_names = list(state.branches.keys())

        ntemps, nwalkers, _, _ = state.branches[all_branch_names[0]].shape

        # get gibbs sampling information
        for branch_names_run, inds_run in self.gibbs_sampling_setup_iterator(
            all_branch_names
        ):
            # setup proposals based on Gibbs sampling
            (
                coords_going_for_proposal,
                inds_going_for_proposal,
                at_least_one_proposal,
            ) = self.setup_proposals(
                branch_names_run, inds_run, state.branches_coords, state.branches_inds
            )

            if not at_least_one_proposal:
                continue

            # prepare accepted fraction
            accepted_here = np.zeros((ntemps, nwalkers), dtype=bool)
            for split in range(self.nsplits):
                # get split information
                S1 = inds == split
                num_total_here = np.sum(inds == split)
                nwalkers_here = np.sum(S1[0])

                all_inds_shaped = all_inds[S1].reshape(ntemps, nwalkers_here)

                # inds including gibbs information
                new_inds = {
                    name: np.take_along_axis(
                        state.branches[name].inds,
                        all_inds_shaped[:, :, None],
                        axis=1,
                    )
                    for name in state.branches
                }

                # the actual inds for the subset
                real_inds_subset = {
                    name: new_inds[name] for name in inds_going_for_proposal
                }

                # actual coordinates of subset
                temp_coords = {
                    name: np.take_along_axis(
                        state.branches_coords[name],
                        all_inds_shaped[:, :, None, None],
                        axis=1,
                    )
                    for name in state.branches_coords
                }

                # prepare the sets for each model
                # goes into the proposal as (ntemps * (nwalkers / subset size), nleaves_max, ndim)
                sets = {
                    key: [
                        np.take_along_axis(
                            state.branches[key].coords,
                            all_inds[inds == j].reshape(ntemps, -1)[:, :, None, None],
                            axis=1,
                        )
                        for j in range(self.nsplits)
                    ]
                    for key in branch_names_run
                }

                # setup s and c based on splits
                s = {key: sets[key][split] for key in sets}
                c = {key: sets[key][:split] + sets[key][split + 1 :] for key in sets}

                # need to trick stretch proposal into using the dimenionality associated
                # with Gibbs sampling if it is being used
                gibbs_ndim = 0
                for brn, ir in zip(branch_names_run, inds_run):
                    if ir is not None:
                        gibbs_ndim += ir.sum()
                    else:
                        # nleaves * ndim
                        gibbs_ndim += np.prod(state.branches[brn].shape[-2:])

                # Get the move-specific proposal.
                q, factors = self.get_proposal(
                    s, c, model.random, gibbs_ndim=gibbs_ndim
                )

                # account for gibbs sampling
                self.cleanup_proposals_gibbs(branch_names_run, inds_run, q, temp_coords)

                # setup supplemental information
                if state.supplemental is not None:
                    # TODO: should there be a copy?
                    new_supps = BranchSupplemental(
                        state.supplemental.take_along_axis(all_inds_shaped, axis=1),
                        base_shape=(ntemps, nwalkers),
                        copy=False,
                    )

                else:
                    new_supps = None

                # default for removing inds info from supp
                if not np.all(
                    np.asarray(list(state.branches_supplemental.values())) == None
                ):
                    new_branch_supps_tmp = {
                        name: state.branches[name].branch_supplemental.take_along_axis(
                            all_inds_shaped[:, :, None], axis=1
                        )
                        for name in state.branches
                        if state.branches[name].branch_supplemental is not None
                    }

                    new_branch_supps = {
                        name: BranchSupplemental(
                            new_branch_supps_tmp[name],
                            base_shape=new_inds[name].shape,
                            copy=False,
                        )
                        for name in new_branch_supps_tmp
                    }

                else:
                    new_branch_supps = None

                # order everything properly
                q, new_inds, new_branch_supps = self.ensure_ordering(
                    list(state.branches.keys()), q, new_inds, new_branch_supps
                )

                # Compute prior of the proposed position
                # new_inds_prior is adjusted if product-space is used
                logp = model.compute_log_prior_fn(
                    q,
                    inds=new_inds,
                    supps=new_supps,
                    branch_supps=new_branch_supps,
                )

                self.fix_logp_gibbs(branch_names_run, inds_run, logp, real_inds_subset)

                # Compute the lnprobs of the proposed position.
                logl, new_blobs = model.compute_log_like_fn(
                    q,
                    inds=new_inds,
                    logp=logp,
                    supps=new_supps,
                    branch_supps=new_branch_supps,
                )

                # catch and warn about nans
                if np.any(np.isnan(logl)):
                    logl[np.isnan(logl)] = -1e300
                    warnings.warn("Getting Nan in likelihood computation.")

                logP = self.compute_log_posterior(logl, logp)

                prev_logl = np.take_along_axis(state.log_like, all_inds_shaped, axis=1)

                prev_logp = np.take_along_axis(state.log_prior, all_inds_shaped, axis=1)

                # takes care of tempering
                prev_logP = self.compute_log_posterior(prev_logl, prev_logp)

                lnpdiff = factors + logP - prev_logP

                keep = lnpdiff > np.log(model.random.rand(ntemps, nwalkers_here))

                # if gibbs sampling, this will say it is accepted if
                # any of the gibbs proposals were accepted
                np.put_along_axis(
                    accepted_here,
                    all_inds_shaped,
                    keep,
                    axis=1,
                )

                # readout for total
                accepted = (accepted.astype(int) + accepted_here.astype(int)).astype(
                    bool
                )
                # new state
                new_state = State(
                    q,
                    log_like=logl,
                    log_prior=logp,
                    blobs=new_blobs,
                    inds=new_inds,
                    supplemental=new_supps,
                    branch_supplemental=new_branch_supps,
                )

                # update state
                state = self.update(
                    state, new_state, accepted_here, subset=all_inds_shaped
                )

            # add to move-specific accepted information
            self.accepted += accepted
            self.num_proposals += 1

        # temp swaps
        if self.temperature_control is not None:
            state = self.temperature_control.temper_comps(state)

        return state, accepted



from ast import Import
import numpy as np

try:
    import cupy as xp

except (ModuleNotFoundError, ImportError) as e:
    pass


class PeriodicContainer:
    """Perform operations for periodic parameters

    Args:
        periodic_in (dict): Keys are ``branch_names``. Values are
            dictionaries. These dictionaries have keys as the parameter
            indexes and values their associated period.

    """

    def __init__(self, periodic, key_order=None):

        # store all the information
        self.periodic = periodic
        inds_periodic = {}
        periods = {}
        for key in periodic:
            if periodic[key] is None:
                continue
            inds_periodic[key] = []
            periods[key] = []
            for var, period in periodic[key].items():
                if isinstance(var, str):
                    if key_order is None:
                        raise ValueError(f"If providing str values for the variable names, must provide key_order argument.")

                    index = key_order[key].index(var)
                    inds_periodic[key].append(index)
                    periods[key].append(period)

        self.inds_periodic = {
            key: np.asarray(tmp) for key, tmp in inds_periodic.items()
        }
        self.periods = {
            key: np.asarray(tmp) for key, tmp in periods.items()
        }

    def distance(self, p1, p2, xp=None):
        """Move from p1 to p2 with periodic distance control

        Args:
            p1 (dict): If dict, keys are ``branch_names``
                and values are positions with parameters along the final dimension.
            p2 (dict): If dict, keys are ``branch_names``
                and values are positions with parameters along the final dimension.
            xp (object, optional): ``numpy`` or ``cupy``. If ``None``, use ``numpy``.
                (default: ``None``)

        Returns:
            dict: Distances accounting for periodicity.
                    Keys are branch names and values are distance arrays.

        """

        # cupy or numpy
        if xp is None:
            xp = np

        # make sure both have same branches
        assert list(p1.keys()) == list(p2.keys())

        names = list(p1.keys())

        # prepare output
        out_diff = {}
        for key in names:

            # get basic distance
            diff = p2[key] - p1[key]

            # no periodic parameters for this key
            if key not in self.periods:
                out_diff[key] = diff
                continue

            # get period info
            periods = xp.asarray(self.periods[key])
            inds_periodic = xp.asarray(self.inds_periodic[key])

            if len(self.periods[key]) > 0:
                # get specific periodic parameterss
                diff_periodic = diff[:, :, inds_periodic]

                # fix when the distance is over 1/2 period away
                inds_fix = (
                    xp.abs(diff_periodic) > periods[xp.newaxis, xp.newaxis, :] / 2.0
                )

                # wrap back to make proper periodic distance
                new_s = -(
                    periods[xp.newaxis, xp.newaxis, :] - p1[key][:, :, inds_periodic]
                ) * (diff_periodic < 0.0) + (
                    periods[xp.newaxis, xp.newaxis, :] + p1[key][:, :, inds_periodic]
                ) * (
                    diff_periodic >= 0.0
                )

                # fill new information
                diff_periodic[inds_fix] = (
                    p2[key][:, :, inds_periodic][inds_fix] - new_s[inds_fix]
                )
                diff[:, :, inds_periodic] = diff_periodic

            out_diff[key] = diff

        return out_diff

    def wrap(self, p, xp=None):
        """Wrap p with periodic distance control

        Args:
            p (dict): If dict, keys are ``branch_names``
                and values are positions with parameters along the final dimension.
            xp (object, optional): ``numpy`` or ``cupy``. If ``None``, use ``numpy``.
                (default: ``None``)

        """

        # cupy or numpy
        if xp is None:
            xp = np

        names = list(p.keys())
        # wrap for each branch
        for key in names:
            pos = p[key]

            if len(self.periods[key]) > 0:
                # get periodic information
                periods = xp.asarray(self.periods[key])
                inds_periodic = xp.asarray(self.inds_periodic[key])
                # wrap
                pos[:, :, inds_periodic] = (
                    pos[:, :, inds_periodic] % periods[xp.newaxis, xp.newaxis, :]
                )

            # fill new info
            p[key] = pos

        return p



"""GPU Backend Tools."""

# ruff: noqa: E402
try:
    from eryn._version import (  # pylint: disable=E0401,E0611
        __version__,
        __version_tuple__,
    )

except ModuleNotFoundError:
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

    try:
        __version__ = version(__name__)
        __version_tuple__ = tuple(__version__.split("."))
    except PackageNotFoundError:  # pragma: no cover
        __version__ = "unknown"
        __version_tuple__ = (0, 0, 0, "unknown")
    finally:
        del version, PackageNotFoundError

_is_editable: bool
try:
    from . import _editable

    _is_editable = True
    del _editable
except (ModuleNotFoundError, ImportError):
    _is_editable = False

__all__ = [
    "__version__",
    "__version_tuple__",
    "_is_editable",
]



# -*- coding: utf-8 -*-

# from .plot import PlotContainer
from .utility import *
from .periodic import *
from .transform import *
from .updates import *
from .stopping import *

__all__ = ["PlotContainer", "PeriodicContainer", "TransformContainer"]



# -*- coding: utf-8 -*-

from .backend import Backend

from .hdfbackend import HDFBackend, TempHDFBackend

__all__ = ["Backend", "HDFBackend", "TempHDFBackend", "get_test_backends"]


def get_test_backends():
    backends = [Backend]

    try:
        import h5py  # NOQA
    except ImportError:
        pass
    else:
        backends.append(TempHDFBackend)

    return backends



import numpy as np

from eryn.moves.multipletry import MultipleTryMoveRJ
from eryn.moves import DistributionGenerateRJ


class MTDistGenMoveRJ(MultipleTryMoveRJ, DistributionGenerateRJ):
    def __init__(self, generate_dist, *args, **kwargs):
        """Perform a reversible-jump multiple try move based on a distribution.
    
        Distribution must be independent of the current point.
        
        This is effectively an example of the mutliple try class inheritance structure.
        
        Args:
            generate_dist (dict): Keys are branch names and values are :class:`ProbDistContainer` objects 
                that have ``logpdf`` and ``rvs`` methods. If you 
            *args (tuple, optional): Additional arguments to pass to parent classes.
            **kwargs (dict, optional): Keyword arguments passed to parent classes.

            
        """
        kwargs["rj"] = True
        MultipleTryMoveRJ.__init__(self, **kwargs)
        DistributionGenerateRJ.__init__(self, generate_dist, *args, **kwargs)

        self.generate_dist = generate_dist

    def special_generate_logpdf(self, generated_coords):
        """Get logpdf of generated coordinates.
        
        Args:
            generated_coords (np.ndarray): Current coordinates of walkers. 
            
        Returns:
            np.ndarray: logpdf of generated points.
            """
        return self.generate_dist[self.key_in].logpdf(generated_coords)

    def special_generate_func(
        self, coords, random, size=1, fill_tuple=None, fill_values=None
    ):
        """Generate samples and calculate the logpdf of their proposal function.
        
        Args:
            coords (np.ndarray): Current coordinates of walkers. 
            random (obj): Random generator.
            *args (tuple, optional): additional arguments passed by overwriting the 
                ``get_proposal`` function and passing ``args_generate`` keyword argument.
            size (int, optional): Number of tries to generate. 
            fill_tuple (tuple, optional): Length 2 tuple with the indexing of which values to fill
                when generating. Can be used for auxillary proposals or reverse RJ proposals. First index is the index into walkers and the second index is 
                the index into the number of tries. (default: ``None``)
            fill_values (np.ndarray): values to fill associated with ``fill_tuple``. Should 
                have size ``(len(fill_tuple[0]), ndim)``. (default: ``None``).
            **kwargs (tuple, optional): additional keyword arguments passed by overwriting the 
                ``get_proposal`` function and passing ``kwargs_generate`` keyword argument.

        Returns:
            tuple: (generated points, logpdf of generated points).
        
        """
        nwalkers = coords.shape[0]

        if not isinstance(size, int):
            raise ValueError("size must be an int.")

        generated_coords = self.generate_dist[self.key_in].rvs(size=(nwalkers, size))

        if fill_values is not None:
            generated_coords[fill_tuple] = fill_values

        generated_logpdf = self.special_generate_logpdf(
            generated_coords.reshape(nwalkers * size, -1)
        ).reshape(nwalkers, size)

        return generated_coords, generated_logpdf

    def set_coords_and_inds(self, generated_coords, inds_leaves_rj=None):
        """Setup coordinates for prior and Likelihood
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``.
            inds_leaves_rj (np.ndarray): Index into each individual walker giving the
                leaf index associated with this proposal. Should only be used if ``self.rj is True``. (default: ``None``)

        Returns:
            dict: Coordinates for Likelihood and Prior.
        
        """
        coords_in = np.repeat(
            self.current_state.branches[self.key_in].coords.reshape(
                (1, -1,) + self.current_state.branches[self.key_in].coords.shape[-2:]
            ),
            self.num_try,
            axis=1,
        )
        coords_in[
            (
                np.zeros(coords_in.shape[0], dtype=int),
                np.arange(coords_in.shape[1]),
                np.repeat(inds_leaves_rj, self.num_try),
            )
        ] = generated_coords.reshape(-1, coords_in.shape[-1])
        inds_in = np.repeat(
            self.current_state.branches[self.key_in].inds.reshape(
                (1, -1,) + self.current_state.branches[self.key_in].inds.shape[-1:]
            ),
            self.num_try,
            axis=1,
        )
        inds_in[
            (
                np.zeros(coords_in.shape[0], dtype=int),
                np.arange(inds_in.shape[1]),
                np.repeat(inds_leaves_rj, self.num_try),
            )
        ] = True

        coords_in_dict = {}
        inds_in_dict = {}
        for key in self.current_state.branches.keys():
            if key == self.key_in:
                coords_in_dict[key] = coords_in
                inds_in_dict[key] = inds_in

            else:
                coords_in_dict[key] = self.current_state.branches[key].coords.reshape(
                    (1, -1) + self.current_state.branches[key].shape[-2:]
                )
                # expand to express multiple tries
                coords_in_dict[key] = np.tile(
                    coords_in_dict[key], (1, self.num_try, 1)
                ).reshape(
                    coords_in_dict[key].shape[0],
                    coords_in_dict[key].shape[1] * self.num_try,
                    coords_in_dict[key].shape[2],
                    coords_in_dict[key].shape[3],
                )
                inds_in_dict[key] = self.current_state.branches[key].inds.reshape(
                    (1, -1) + self.current_state.branches[key].shape[-2:-1]
                )
                # expand to express multiple tries
                inds_in_dict[key] = np.tile(
                    inds_in_dict[key], (1, self.num_try)
                ).reshape(
                    inds_in_dict[key].shape[0],
                    inds_in_dict[key].shape[1] * self.num_try,
                    inds_in_dict[key].shape[2],
                )

        return coords_in_dict, inds_in_dict

    def special_like_func(self, generated_coords, inds_leaves_rj=None, **kwargs):
        """Calculate the Likelihood for sampled points.
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``. 
            inds_leaves_rj (np.ndarray): Index into each individual walker giving the
                leaf index associated with this proposal. Should only be used if ``self.rj is True``. (default: ``None``)

        Returns:
            np.ndarray: Likelihood values with shape ``(generated_coords.shape[0], num_try).``
        
        """
        coords_in, inds_in = self.set_coords_and_inds(
            generated_coords, inds_leaves_rj=inds_leaves_rj
        )
        ll = self.current_model.compute_log_like_fn(coords_in, inds=inds_in)[0]
        ll = ll[0].reshape(-1, self.num_try)
        return ll

    def special_prior_func(self, generated_coords, inds_leaves_rj=None, **kwargs):
        """Calculate the Prior for sampled points.
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``. 
            inds_leaves_rj (np.ndarray): Index into each individual walker giving the
                leaf index associated with this proposal. Should only be used if ``self.rj is True``. (default: ``None``)
            
        Returns:
            np.ndarray: Prior values with shape ``(generated_coords.shape[0], num_try).``
        
        """
        coords_in, inds_in = self.set_coords_and_inds(
            generated_coords, inds_leaves_rj=inds_leaves_rj
        )
        lp = self.current_model.compute_log_prior_fn(coords_in, inds=inds_in)
        return lp.reshape(-1, self.num_try)




# -*- coding: utf-8 -*-

import numpy as np
from ..state import State
from copy import deepcopy

__all__ = ["TemperatureControl"]


def make_ladder(ndim, ntemps=None, Tmax=None):
    """
    Returns a ladder of :math:`\\beta \\equiv 1/T` under a geometric spacing that is determined by the
    arguments ``ntemps`` and ``Tmax``.  The temperature selection algorithm works as follows:
    Ideally, ``Tmax`` should be specified such that the tempered posterior looks like the prior at
    this temperature.  If using adaptive parallel tempering, per `arXiv:1501.05823
    <http://arxiv.org/abs/1501.05823>`_, choosing ``Tmax = inf`` is a safe bet, so long as
    ``ntemps`` is also specified.

    This function is originally from ``ptemcee`` `github.com/willvousden/ptemcee <https://github.com/willvousden/ptemcee>`_.

    Temperatures are chosen according to the following algorithm:
    * If neither ``ntemps`` nor ``Tmax`` is specified, raise an exception (insufficient
    information).
    * If ``ntemps`` is specified but not ``Tmax``, return a ladder spaced so that a Gaussian
    posterior would have a 25% temperature swap acceptance ratio.
    * If ``Tmax`` is specified but not ``ntemps``:
    * If ``Tmax = inf``, raise an exception (insufficient information).
    * Else, space chains geometrically as above (for 25% acceptance) until ``Tmax`` is reached.
    * If ``Tmax`` and ``ntemps`` are specified:
    * If ``Tmax = inf``, place one chain at ``inf`` and ``ntemps-1`` in a 25% geometric spacing.
    * Else, use the unique geometric spacing defined by ``ntemps`` and ``Tmax``.`

    Args:
        ndim (int): The number of dimensions in the parameter space.
        ntemps (int, optional): If set, the number of temperatures to generate.
        Tmax (float, optional): If set, the maximum temperature for the ladder.

    Returns:
        np.ndarray[ntemps]: Output inverse temperature (beta) array.

    Raises:
        ValueError: Improper inputs.

    """

    # make sure all inputs are okay
    if type(ndim) != int or ndim < 1:
        raise ValueError("Invalid number of dimensions specified.")
    if ntemps is None and Tmax is None:
        raise ValueError("Must specify one of ``ntemps`` and ``Tmax``.")
    if Tmax is not None and Tmax <= 1:
        raise ValueError("``Tmax`` must be greater than 1.")
    if ntemps is not None and (type(ntemps) != int or ntemps < 1):
        raise ValueError("Invalid number of temperatures specified.")

    # step size in temperature based on ndim
    tstep = np.array(
        [
            25.2741,
            7.0,
            4.47502,
            3.5236,
            3.0232,
            2.71225,
            2.49879,
            2.34226,
            2.22198,
            2.12628,
            2.04807,
            1.98276,
            1.92728,
            1.87946,
            1.83774,
            1.80096,
            1.76826,
            1.73895,
            1.7125,
            1.68849,
            1.66657,
            1.64647,
            1.62795,
            1.61083,
            1.59494,
            1.58014,
            1.56632,
            1.55338,
            1.54123,
            1.5298,
            1.51901,
            1.50881,
            1.49916,
            1.49,
            1.4813,
            1.47302,
            1.46512,
            1.45759,
            1.45039,
            1.4435,
            1.4369,
            1.43056,
            1.42448,
            1.41864,
            1.41302,
            1.40761,
            1.40239,
            1.39736,
            1.3925,
            1.38781,
            1.38327,
            1.37888,
            1.37463,
            1.37051,
            1.36652,
            1.36265,
            1.35889,
            1.35524,
            1.3517,
            1.34825,
            1.3449,
            1.34164,
            1.33847,
            1.33538,
            1.33236,
            1.32943,
            1.32656,
            1.32377,
            1.32104,
            1.31838,
            1.31578,
            1.31325,
            1.31076,
            1.30834,
            1.30596,
            1.30364,
            1.30137,
            1.29915,
            1.29697,
            1.29484,
            1.29275,
            1.29071,
            1.2887,
            1.28673,
            1.2848,
            1.28291,
            1.28106,
            1.27923,
            1.27745,
            1.27569,
            1.27397,
            1.27227,
            1.27061,
            1.26898,
            1.26737,
            1.26579,
            1.26424,
            1.26271,
            1.26121,
            1.25973,
        ]
    )

    if ndim > tstep.shape[0]:
        # An approximation to the temperature step at large
        # dimension
        tstep = 1.0 + 2.0 * np.sqrt(np.log(4.0)) / np.sqrt(ndim)
    else:
        # get correct step for dimension
        tstep = tstep[ndim - 1]

    # wheter to add the infinite temperature to the end
    appendInf = False
    if Tmax == np.inf:
        appendInf = True
        Tmax = None
        # non-infinite temperatures will now have 1 less
        ntemps = ntemps - 1

    if ntemps is not None:
        if Tmax is None:
            # Determine Tmax from ntemps.
            Tmax = tstep ** (ntemps - 1)
    else:
        if Tmax is None:
            raise ValueError(
                "Must specify at least one of ``ntemps" " and " "finite ``Tmax``."
            )

        # Determine ntemps from Tmax.
        ntemps = int(np.log(Tmax) / np.log(tstep) + 2)

    betas = np.logspace(0, -np.log10(Tmax), ntemps)
    if appendInf:
        # Use a geometric spacing, but replace the top-most temperature with
        # infinity.
        betas = np.concatenate((betas, [0]))

    return betas


class TemperatureControl(object):
    """Controls the temperature ladder and operations in the sampler.

    All of the tempering features within Eryn are controlled from this class.
    This includes the evaluation of the tempered posterior, swapping between temperatures, and
    the adaptation of the temperatures over time. The adaptive tempering model can be
    found in the Eryn paper as well as the paper for `ptemcee`, which acted
    as a basis for the code below.

    Args:
        effective_ndim (int): Effective dimension used to determine temperatures if betas not given.
        nwalkers (int): Number of walkers in the sampler. Must maintain proper order of branches.
        ntemps (int, optional): Number of temperatures. If this is provided rather than ``betas``,
            :func:`make_ladder` will be used to generate the temperature ladder. (default: 1)
        betas (np.ndarray[ntemps], optional): If provided, will use as the array of inverse temperatures.
            (default: ``None``).
        Tmax (float, optional): If provided and ``betas`` is not provided, this will be included with
            ``ntemps`` when determing the temperature ladder with :func:`make_ladder`.
            See that functions docs for more information. (default: ``None``)
        adaptive (bool, optional): If ``True``, adapt the temperature ladder during sampling.
            (default: ``True``).
        adaptation_lag (int, optional): lag parameter from
            `arXiv:1501.05823 <http://arxiv.org/abs/1501.05823>`_. ``adaptation_lag`` must be
            much greater than ``adapation_time``. (default: 10000)
        adaptation_time (int, optional): initial amplitude of adjustments from
            `arXiv:1501.05823 <http://arxiv.org/abs/1501.05823>`_. ``adaptation_lag`` must be
            much greater than ``adapation_time``. (default: 100)
        stop_adaptation (int, optional): If ``stop_adaptation > 0``, the adapating will stop after
            ``stop_adaption`` steps. The number of steps is counted as the number times adaptation
            has happened which is generally once per sampler iteration. For example,
            if you only want to adapt temperatures during burn-in, you set ``stop_adaption = burn``.
            This can become complicated when using the repeating proposal options, so the
            user must be careful and verify constant temperatures in the backend.
            (default: -1)
        permute (bool, optional): If ``True``, permute the walkers in each temperature during
            swaps. (default: ``True``)
        skip_swap_supp_names (list, optional): List of strings that indicate supplemental keys that are not to be swapped.
            (default: ``[]``)


    """

    def __init__(
        self,
        effective_ndim,
        nwalkers,
        ntemps=1,
        betas=None,
        Tmax=None,
        adaptive=True,
        adaptation_lag=10000,
        adaptation_time=100,
        stop_adaptation=-1,
        permute=True,
        skip_swap_supp_names=[],
    ):

        if betas is None:
            if ntemps == 1:
                betas = np.array([1.0])
            else:
                # A compromise for building a temperature ladder for the case of rj.
                # We start by assuming that the dimensionality will be defined by the number of
                # components. We take that maximum divided by two, and multiply it with the higher
                # dimensional component.
                betas = make_ladder(effective_ndim, ntemps=ntemps, Tmax=Tmax)

        # store information
        self.nwalkers = nwalkers
        self.betas = betas
        self.ntemps = ntemps = len(betas)
        self.permute = permute
        self.skip_swap_supp_names = skip_swap_supp_names

        # number of times adapted
        self.time = 0

        # store adapting inf
        self.adaptive = adaptive
        self.adaptation_time, self.adaptation_lag = adaptation_time, adaptation_lag
        self.stop_adaptation = stop_adaptation

        self.swaps_proposed = np.full(self.ntemps - 1, self.nwalkers)

    def compute_log_posterior_tempered(self, logl, logp, betas=None):
        """Compute the log of the tempered posterior

        Args:
            logl (np.ndarray): Log of the Likelihood. Can be 1D or 2D array. If 2D,
                must have shape ``(ntemps, nwalkers)``. If 1D, ``betas`` must be provided
                with the same shape.
            logp (np.ndarray): Log of the Prior. Can be 1D or 2D array. If 2D,
                must have shape ``(ntemps, nwalkers)``. If 1D, ``betas`` must be provided
                with the same shape.
            betas (np.ndarray[ntemps]): If provided, inverse temperatures as 1D array.
                If not provided, it will use ``self.betas``. (default: ``None``)

        Returns:
            np.ndarray: Log of the temperated posterior.

        Raises:
            AssertionError: Inputs are incorrectly shaped.

        """
        assert logl.shape == logp.shape
        tempered_logl = self.tempered_likelihood(logl, betas=betas)
        return tempered_logl + logp

    def tempered_likelihood(self, logl, betas=None):
        """Compute the log of the tempered Likelihood

        From `ptemcee`: "This is usually a mundane multiplication, except for the special case where
        beta == 0 *and* we're outside the likelihood support.
        Here, we find a singularity that demands more careful attention; we allow the
        likelihood to dominate the temperature, since wandering outside the
        likelihood support causes a discontinuity."

        Args:
            logl (np.ndarray): Log of the Likelihood. Can be 1D or 2D array. If 2D,
                must have shape ``(ntemps, nwalkers)``. If 1D, ``betas`` must be provided
                with the same shape.
            betas (np.ndarray[ntemps]): If provided, inverse temperatures as 1D array.
                If not provided, it will use ``self.betas``. (default: ``None``)

        Returns:
            np.ndarray: Log of the temperated Likelihood.

        Raises:
            ValueError: betas not provided if needed.

        """
        # perform calculation on 1D likelihoods.
        if logl.ndim == 1:
            if betas is None:
                raise ValueError(
                    "If inputing a 1D logl array, need to provide 1D betas array of the same length."
                )
            loglT = logl * betas

        else:
            if betas is None:
                betas = self.betas

            with np.errstate(invalid="ignore"):
                loglT = logl * betas[:, None]

        # anywhere the likelihood is nan, turn into -infinity
        loglT[np.isnan(loglT)] = -np.inf

        return loglT

    def do_swaps_indexing(
        self,
        i,
        iperm_sel,
        i1perm_sel,
        dbeta,
        x,
        logP,
        logl,
        logp,
        inds=None,
        blobs=None,
        supps=None,
        branch_supps=None,
    ):

        # for x and inds, just do full copy
        x_temp = {name: np.copy(x[name]) for name in x}
        if inds is not None:
            inds_temp = {name: np.copy(inds[name]) for name in inds}
        if branch_supps is not None:
            branch_supps_temp = {
                name: deepcopy(branch_supps[name]) for name in branch_supps
            }

        logl_temp = np.copy(logl[i, iperm_sel])
        logp_temp = np.copy(logp[i, iperm_sel])
        logP_temp = np.copy(logP[i, iperm_sel])
        if blobs is not None:
            blobs_temp = np.copy(blobs[i, iperm_sel])
        if supps is not None:
            supps_temp = deepcopy(supps[i, iperm_sel])

        # swap from i1 to i
        for name in x:
            # coords first
            x[name][i, iperm_sel, :, :] = x[name][i - 1, i1perm_sel, :, :]

            # then inds
            if inds is not None:
                inds[name][i, iperm_sel, :] = inds[name][i - 1, i1perm_sel, :]

            # do something special for branch_supps in case in contains a large amount of data
            # that is heavy to copy
            if branch_supps[name] is not None:
                tmp = branch_supps[name][i - 1, i1perm_sel, :]

                for key in self.skip_swap_supp_names:
                    tmp.pop(key)

                branch_supps[name][i, iperm_sel, :] = tmp
                """# where the inds are alive in the current permutation
                # need inds_temp because that is the original
                inds_i = np.where(inds_temp[name][i][iperm_sel])

                # gives the associated walker for each spot in the permuted array
                walker_inds_i = iperm_sel[inds_i[0]]

                # represents which permuted leaves are alive
                leaf_inds_i = inds_i[1]

                # all of these are at the same temperature
                temp_inds_i = np.full_like(leaf_inds_i, i)

                # repeat all for the i1 permutated temperature
                # need inds_temp because that is the original
                inds_i1 = np.where(inds_temp[name][i - 1][i1perm_sel])
                walker_inds_i1 = i1perm_sel[inds_i1[0]]
                leaf_inds_i1 = inds_i1[1]
                temp_inds_i1 = np.full_like(leaf_inds_i1, i - 1)

                # go through the values within each branch supplemental holder
                # do direct movement of things that need to change
                # rather than copying the whole thing
                for name2 in branch_supps[name].holder:
                    # store temperarily
                    bring_back_branch_supps = (
                        branch_supps[name]
                        .holder[name2][(temp_inds_i, walker_inds_i, leaf_inds_i)]
                        .copy()
                    )

                    # make switch from i1 to i
                    branch_supps[name].holder[name2][
                        (temp_inds_i, walker_inds_i, leaf_inds_i)
                    ] = branch_supps[name].holder[name2][
                        (temp_inds_i1, walker_inds_i1, leaf_inds_i1)
                    ]

                    # make switch from i to i1
                    branch_supps[name].holder[name2][
                        (temp_inds_i1, walker_inds_i1, leaf_inds_i1)
                    ] = bring_back_branch_supps"""

        # switch everythin else from i1 to i
        logl[i, iperm_sel] = logl[i - 1, i1perm_sel]
        logp[i, iperm_sel] = logp[i - 1, i1perm_sel]
        logP[i, iperm_sel] = logP[i - 1, i1perm_sel] - dbeta * logl[i - 1, i1perm_sel]
        if blobs is not None:
            blobs[i, iperm_sel] = blobs[i - 1, i1perm_sel]
        if supps is not None:
            tmp_supps = supps[i - 1, i1perm_sel]
            for key in self.skip_swap_supp_names:
                tmp_supps.pop(key)
            supps[i, iperm_sel] = tmp_supps

        # switch x from i to i1
        for name in x:
            x[name][i - 1, i1perm_sel, :, :] = x_temp[name][i, iperm_sel, :, :]
            if inds is not None:
                inds[name][i - 1, i1perm_sel, :] = inds_temp[name][i, iperm_sel, :]
            if branch_supps[name] is not None:
                tmp = branch_supps_temp[name][i, iperm_sel, :]

                for key in self.skip_swap_supp_names:
                    tmp.pop(key)
                branch_supps[name][i - 1, i1perm_sel, :] = tmp

        # switch the rest from i to i1
        logl[i - 1, i1perm_sel] = logl_temp
        logp[i - 1, i1perm_sel] = logp_temp
        logP[i - 1, i1perm_sel] = logP_temp + dbeta * logl_temp

        if blobs is not None:
            blobs[i - 1, i1perm_sel] = blobs_temp
        if supps is not None:
            tmp_supps = supps_temp
            for key in self.skip_swap_supp_names:
                tmp_supps.pop(key)
            supps[i - 1, i1perm_sel] = tmp_supps

        return (x, logP, logl, logp, inds, blobs, supps, branch_supps)

    def temperature_swaps(
        self, x, logP, logl, logp, inds=None, blobs=None, supps=None, branch_supps=None
    ):
        """Perform parallel-tempering temperature swaps

        This function performs the swapping between neighboring temperatures. It cascades from
        high temperature down to low temperature.

        Args:
            x (dict): Dictionary with keys as branch names and values as coordinate arrays.
            logP (np.ndarray[ntemps, nwalkers]): Log of the posterior probability.
            logl (np.ndarray[ntemps, nwalkers]): Log of the Likelihood.
            logp (np.ndarray[ntemps, nwalkers]): Log of the prior probability.
            inds (dict, optional): Dictionary with keys as branch names and values as the index arrays
                indicating which leaves are used. (default: ``None``)
            blobs (object, optional): Blobs associated with each walker. (default: ``None``)
            supps (object, optional): :class:`eryn.state.BranchSupplemental` object. (default: ``None``)
            branch_supps (dict, optional): Dictionary with keys as branch names and values as
                :class:`eryn.state.BranchSupplemental` objects for each branch (can be ``None`` for some branches). (default: ``None``)

        Returns:
            tuple: All of the information that was input now swapped (output in the same order as input).

        """

        ntemps, nwalkers = self.ntemps, self.nwalkers

        # prepare information on how many swaps are accepted this time
        self.swaps_accepted = np.empty(ntemps - 1)

        # iterate from highest to lowest temperatures
        for i in range(ntemps - 1, 0, -1):

            # get both temperature rungs
            bi = self.betas[i]
            bi1 = self.betas[i - 1]

            # difference in inverse temps
            dbeta = bi1 - bi

            # permute the indices for the walkers in each temperature to randomize swap positions
            if self.permute:
                iperm = np.random.permutation(nwalkers)
                i1perm = np.random.permutation(nwalkers)

            # do not permute if desired
            else:
                iperm = np.arange(nwalkers)
                i1perm = np.arange(nwalkers)

            # random draw that produces log of the acceptance fraction
            raccept = np.log(np.random.uniform(size=nwalkers))

            # log of the detailed balance fraction
            paccept = dbeta * (logl[i, iperm] - logl[i - 1, i1perm])

            # How many swaps were accepted
            sel = paccept > raccept
            self.swaps_accepted[i - 1] = np.sum(sel)

            (x, logP, logl, logp, inds, blobs, supps, branch_supps) = (
                self.do_swaps_indexing(
                    i,
                    iperm[sel],
                    i1perm[sel],
                    dbeta,
                    x,
                    logP,
                    logl,
                    logp,
                    inds=inds,
                    blobs=blobs,
                    supps=supps,
                    branch_supps=branch_supps,
                )
            )

        return (x, logP, logl, logp, inds, blobs, supps, branch_supps)

    def _get_ladder_adjustment(self, time, betas0, ratios):
        """
        Execute temperature adjustment according to dynamics outlined in
        `arXiv:1501.05823 <http://arxiv.org/abs/1501.05823>`_.
        """
        betas = betas0.copy()

        # Modulate temperature adjustments with a hyperbolic decay.
        decay = self.adaptation_lag / (time + self.adaptation_lag)
        kappa = decay / self.adaptation_time

        # Construct temperature adjustments.
        dSs = kappa * (ratios[:-1] - ratios[1:])

        # Compute new ladder (hottest and coldest chains don't move).
        deltaTs = np.diff(1 / betas[:-1])
        deltaTs *= np.exp(dSs)
        betas[1:-1] = 1 / (np.cumsum(deltaTs) + 1 / betas[0])

        # Don't mutate the ladder here; let the client code do that.
        return betas - betas0

    def adapt_temps(self):
        # determine ratios of swaps accepted to swaps proposed (the ladder is fixed)
        ratios = self.swaps_accepted / self.swaps_proposed

        # adapt if desired
        if self.adaptive and self.ntemps > 1:
            if self.stop_adaptation < 0 or self.time < self.stop_adaptation:
                dbetas = self._get_ladder_adjustment(self.time, self.betas, ratios)
                self.betas += dbetas

            # only increase time if it is adaptive.
            self.time += 1

    def temper_comps(self, state, adapt=True):
        """Perfrom temperature-related operations on a state.

        This includes making swaps and then adapting the temperatures for the next round.

        Args:
            state (object): Filled ``State`` object.
            adapt (bool, optional): If True, swaps are to be performed, but no
                adaptation is made. In this case, ``self.time`` does not increase by 1.
                (default: ``True``)

        Returns:
            :class:`eryn.state.State`: State object after swaps.

        """
        # get initial values
        logl = state.log_like
        logp = state.log_prior

        # do posterior just for the hell of it
        logP = self.compute_log_posterior_tempered(logl, logp)

        # make swaps
        x, logP, logl, logp, inds, blobs, supps, branch_supps = self.temperature_swaps(
            state.branches_coords,
            logP.copy(),
            logl.copy(),
            logp.copy(),
            inds=state.branches_inds,
            blobs=state.blobs,
            supps=state.supplemental,
            branch_supps=state.branches_supplemental,
        )

        if adapt and self.adaptive and self.ntemps > 1:
            self.adapt_temps()

        # create a new state out of the swapped information
        # TODO: make this more memory efficient?
        new_state = State(
            x,
            log_like=logl,
            log_prior=logp,
            blobs=blobs,
            inds=inds,
            betas=self.betas,
            supplemental=supps,
            branch_supplemental=branch_supps,
            random_state=state.random_state,
        )

        return new_state



import numpy as np

from eryn.moves.multipletry import MultipleTryMove
from eryn.moves import MHMove


class MTDistGenMove(MultipleTryMove, MHMove):
    """Perform a multiple try MH move based on a distribution.
    
    Distribution must be independent of the current point.
    
    This is effectively an example of the mutliple try class inheritance structure.

    Args:
        generate_dist (dict): Keys are branch names and values are :class:`ProbDistContainer` objects 
            that have ``logpdf`` and ``rvs`` methods. If you 
        *args (tuple, optional): Additional arguments to pass to parent classes.
        **kwargs (dict, optional): Keyword arguments passed to parent classes.
    
    """

    def __init__(self, generate_dist, **kwargs):

        MultipleTryMove.__init__(self, **kwargs)
        MHMove.__init__(self, **kwargs)

        self.generate_dist = generate_dist

    def special_generate_logpdf(self, generated_coords):
        """Get logpdf of generated coordinates.
        
        Args:
            generated_coords (np.ndarray): Current coordinates of walkers. 
            
        Returns:
            np.ndarray: logpdf of generated points.
            
        """
        return self.generate_dist.logpdf(generated_coords)

    def special_generate_func(
        self, coords, random, size=1, fill_tuple=None, fill_values=None, **kwargs
    ):
        """Generate samples and calculate the logpdf of their proposal function.
        
        Args:
            coords (np.ndarray): Current coordinates of walkers. 
            random (obj): Random generator.
            *args (tuple, optional): additional arguments passed by overwriting the 
                ``get_proposal`` function and passing ``args_generate`` keyword argument.
            size (int, optional): Number of tries to generate. 
            fill_tuple (tuple, optional): Length 2 tuple with the indexing of which values to fill
                when generating. Can be used for auxillary proposals or reverse RJ proposals. First index is the index into walkers and the second index is 
                the index into the number of tries. (default: ``None``)
            fill_values (np.ndarray): values to fill associated with ``fill_tuple``. Should 
                have size ``(len(fill_tuple[0]), ndim)``. (default: ``None``).
            **kwargs (tuple, optional): additional keyword arguments passed by overwriting the 
                ``get_proposal`` function and passing ``kwargs_generate`` keyword argument.

        Returns:
            tuple: (generated points, logpdf of generated points).
        
        """
        nwalkers = coords.shape[0]

        if not isinstance(size, int):
            raise ValueError("size must be an int.")

        # generate coordinates
        generated_coords = self.generate_dist.rvs(size=(nwalkers, size))

        # fill special coordinates
        if fill_values is not None:
            generated_coords[fill_tuple] = fill_values

        # get logpdf
        generated_logpdf = self.special_generate_logpdf(
            generated_coords.reshape(nwalkers * size, -1)
        ).reshape(nwalkers, size)

        return generated_coords, generated_logpdf

    def set_coords_and_inds(self, generated_coords):
        """Setup coordinates for prior and Likelihood
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``. 

        Returns:
            dict: Coordinates for Likelihood and Prior.
        
        """
        ndim = self.current_state.branches[self.key_in].shape[-1]

        coords_in_dict = {}
        for key in self.current_state.branches.keys():
            if key == self.key_in:
                coords_in_dict[key] = generated_coords.reshape(-1, 1, ndim)[None, :]

            else:
                coords_in_dict[key] = self.current_state.branches[key].coords.reshape(
                    (1, -1) + self.current_state.branches[key].shape[-2:]
                )

        return coords_in_dict

    def special_like_func(self, generated_coords, **kwargs):
        """Calculate the Likelihood for sampled points.
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``. 
            **kwargs (dict, optional): For compatibility. 

        Returns:
            np.ndarray: Likelihood values with shape ``(generated_coords.shape[0], num_try).``
        
        """
        coords_in = self.set_coords_and_inds(generated_coords)
        ll = self.current_model.compute_log_like_fn(coords_in)[0]
        ll = ll[0].reshape(-1, self.num_try)
        return ll

    def special_prior_func(self, generated_coords, **kwargs):
        """Calculate the Prior for sampled points.
        
        Args:
            generated_coords (np.ndarray): Generated coordinates with shape ``(number of independent walkers, num_try)``. 
            **kwargs (dict, optional): For compatibility. 
            
        Returns:
            np.ndarray: Prior values with shape ``(generated_coords.shape[0], num_try).``
        
        """
        coords_in = self.set_coords_and_inds(generated_coords)
        lp = self.current_model.compute_log_prior_fn(coords_in)
        return lp.reshape(-1, self.num_try)




# -*- coding: utf-8 -*-

from ..state import BranchSupplemental
import numpy as np

from copy import deepcopy

try:
    import cupy as cp
except (ModuleNotFoundError, ImportError):
    import numpy as cp

__all__ = ["Move"]


class Move(object):
    """Parent class for proposals or "moves"

    Args:
        temperature_control (:class:`tempering.TemperatureControl`, optional):
            This object controls the tempering. It is passed to the parent class
            to moves so that all proposals can share and use temperature settings.
            (default: ``None``)
        periodic (:class:`eryn.utils.PeriodicContainer, optional):
            This object holds periodic information and methods for periodic parameters. It is passed to the parent class
            to moves so that all proposals can share and use periodic information.
            (default: ``None``)
        gibbs_sampling_setup (str, tuple, dict, or list, optional): This sets the Gibbs Sampling setup if
            desired. The Gibbs sampling setup is completely customizable down to the leaf and parameters.
            All of the separate Gibbs sampling splits will be run within 1 call to this proposal.
            If ``None``, run all branches and all parameters. If ``str``, run all parameters within the
            branch given as the string. To enter a branch with a specific set of parameters, you can
            provide a 2-tuple with the first entry as the branch name and the second entry as a 2D
            boolean array of shape ``(nleaves_max, ndim)`` that indicates which leaves and/or parameters
            you want to run. ``None`` can also be entered in the second entry if all parameters are to be run.
            A dictionary is also possible with keys as branch names and values as the same 2D boolean array
            of shape ``(nleaves_max, ndim)`` that indicates which leaves and/or parameters
            you want to run. ``None`` can also be entered in the value of the dictionary
            if all parameters are to be run. If multiple keys are provided in the dictionary, those
            branches will be run simultaneously in the proposal as one iteration of the proposing loop.
            The final option is a list. This is how you make sure to run all the Gibbs splits. Each entry
            of the list can be a string, 2-tuple, or dictionary as described above. The list controls
            the order in which all of these splits are run. (default: ``None``)
        prevent_swaps (bool, optional): If ``True``, do not perform temperature swaps in this move.
        skip_supp_names_update (list, optional): List of names (`str`), that can be in any
            :class:`eryn.state.BranchSupplemental`,
            to skip when updating states (:func:`Move.update`). This is useful if a
            large amount of memory is stored in the branch supplementals.
        is_rj (bool, optional): If using RJ, this should be ``True``. (default: ``False``)
        use_gpu (bool, optional): If ``True``, use ``CuPy`` for computations.
            Use ``NumPy`` if ``use_gpu == False``. (default: ``False``)
        random_seed (int, optional): Set the random seed in ``CuPy/NumPy`` if not ``None``.
            (default: ``None``)

    Raises:
        ValueError: Incorrect inputs.

    Attributes:
        Note: All kwargs are stored as attributes.
        num_proposals (int): the number of times this move has been run. This is needed to
            compute the acceptance fraction.
        gibbs_sampling_setup (list): All of the Gibbs sampling splits as described above.
        xp (obj): ``NumPy`` or ``CuPy``.
        use_gpu (bool): Whether ``Cupy`` (``True``) is used or not (``False``).

    """

    def __init__(
        self,
        temperature_control=None,
        periodic=None,
        gibbs_sampling_setup=None,
        prevent_swaps=False,
        skip_supp_names_update=[],
        is_rj=False,
        use_gpu=False,
        random_seed=None,
        **kwargs
    ):
        # store all information
        self.temperature_control = temperature_control
        self.periodic = periodic
        self.skip_supp_names_update = skip_supp_names_update
        self.prevent_swaps = prevent_swaps

        self._initialize_branch_setup(gibbs_sampling_setup, is_rj=is_rj)

        # keep track of the number of proposals
        self.num_proposals = 0
        self.time = 0

        self.use_gpu = use_gpu

        # set the random seet of the library if desired
        if random_seed is not None:
            self.xp.random.seed(random_seed)

    @property
    def use_gpu(self):
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, use_gpu):
        self._use_gpu = use_gpu

    @property
    def xp(self):
        if self._use_gpu is None:
            raise ValueError("use_gpu has not been set.")
        xp = cp if self.use_gpu else np
        return xp

    def _initialize_branch_setup(self, gibbs_sampling_setup, is_rj=False):
        """Initialize the gibbs setup properly."""
        self.gibbs_sampling_setup = gibbs_sampling_setup

        message_rj = """inputting gibbs indexing at the leaf/parameter level is not allowed 
                                        with an RJ proposal. Only branch names."""

        message_non_rj = """When inputing gibbs indexing and using a 2-tuple, second item must be None or 2D np.ndarray of shape (nleaves_max, ndim)."""

        # setup proposal branches properly
        if self.gibbs_sampling_setup is not None:
            # string indicates one branch (all of it)
            if type(self.gibbs_sampling_setup) not in [str, tuple, list, dict]:
                raise ValueError(
                    "gibbs_sampling_setup must be string, dict, tuple, or list."
                )

            if not isinstance(self.gibbs_sampling_setup, list):
                self.gibbs_sampling_setup = [self.gibbs_sampling_setup]

            gibbs_sampling_setup_tmp = []
            for item in self.gibbs_sampling_setup:
                # all the arguments are treated

                # strings indicate single branch all parameters
                if isinstance(item, str):
                    gibbs_sampling_setup_tmp.append(item)

                # tuple is one branch with a split in the parameters
                elif isinstance(item, tuple):
                    # check inputs
                    assert len(item) == 2
                    if item is not None and is_rj:
                        raise ValueError(message_rj)

                    elif (
                        not isinstance(item[1], np.ndarray) and item[1] is not None
                    ) or (isinstance(item[1], np.ndarray) and item[1].ndim != 2):
                        breakpoint()
                        raise ValueError(message_non_rj)

                    gibbs_sampling_setup_tmp.append(item)

                # dict can include multiple models and parameter splits
                # these will all be in one iteration
                elif isinstance(item, dict):
                    tmp = []
                    for key, value in item.items():
                        # check inputs
                        if value is not None and is_rj:
                            raise ValueError(message_rj)

                        elif (
                            not isinstance(value, np.ndarray) and value is not None
                        ) or (isinstance(value, np.ndarray) and value.ndim != 2):
                            raise ValueError(message_non_rj)

                        tmp.append((key, value))

                    gibbs_sampling_setup_tmp.append(tmp)

                else:
                    raise ValueError(
                        "If providing a list for gibbs_sampling_setup, each item needs to be a string, tuple, or dict."
                    )

            # copy the original for information if needed
            self.gibbs_sampling_setup_input = deepcopy(self.gibbs_sampling_setup)

            # store as the setup that all proposals will follow
            self.gibbs_sampling_setup = gibbs_sampling_setup_tmp

            # now that we have everything out of the input
            # sort into branch names and indices to be run
            branch_names_run_all = []
            inds_run_all = []

            # for each split in the gibbs splits
            for prop_i, proposal_iteration in enumerate(self.gibbs_sampling_setup):
                # break out
                if isinstance(proposal_iteration, tuple):
                    # tuple is 1 entry loop
                    branch_names_run_all.append([proposal_iteration[0]])
                    inds_run_all.append([proposal_iteration[1]])
                elif isinstance(proposal_iteration, str):
                    # string is 1 entry loop
                    branch_names_run_all.append([proposal_iteration])
                    inds_run_all.append([None])

                elif isinstance(proposal_iteration, list):
                    # list allows more branches at the same time
                    branch_names_run_all.append([])
                    inds_run_all.append([])
                    for item in proposal_iteration:
                        if isinstance(item, str):
                            branch_names_run_all[prop_i].append(item)
                            inds_run_all[prop_i].append(None)
                        elif isinstance(item, tuple):
                            branch_names_run_all[prop_i].append(item[0])
                            inds_run_all[prop_i].append(item[1])

            # store information
            self.branch_names_run_all = branch_names_run_all
            self.inds_run_all = inds_run_all

        else:
            # no Gibbs sampling
            self.branch_names_run_all = [None]
            self.inds_run_all = [None]

    def gibbs_sampling_setup_iterator(self, all_branch_names):
        """Iterate through the gibbs splits as a generator

        Args:
            all_branch_names (list): List of all branch names.

        Yields:
            2-tuple: Gibbs sampling split.
                        First entry is the branch names to run and the second entry is the index
                        into the leaves/parameters for this Gibbs split.

        Raises:
            ValueError: Incorrect inputs.

        """
        for branch_names_run, inds_run in zip(
            self.branch_names_run_all, self.inds_run_all
        ):
            # adjust if branch_names_run is None
            if branch_names_run is None:
                branch_names_run = all_branch_names
                inds_run = [None for _ in branch_names_run]
            # yield to the iterator
            yield (branch_names_run, inds_run)

    def setup_proposals(
        self, branch_names_run, inds_run, branches_coords, branches_inds
    ):
        """Setup proposals when gibbs sampling.

        Get inputs into the proposal including Gibbs split information.

        Args:
            branch_names_run (list): List of branch names to run concurrently.
            inds_run (list): List of ``inds`` arrays including Gibbs sampling information.
            branches_coords (dict): Dictionary of coordinate arrays for all branches.
            branches_inds (dict): Dictionary of ``inds`` arrays for all branches.

        Returns:
            tuple:  (coords, inds, at_least_one_proposal)
                        * Coords including Gibbs sampling info.
                        * ``inds`` including Gibbs sampling info.
                        * ``at_least_one_proposal`` is boolean. It is passed out to
                            indicate there is at least one leaf available for the requested branch names.

        """
        inds_going_for_proposal = {}
        coords_going_for_proposal = {}

        at_least_one_proposal = False
        for bnr, ir in zip(branch_names_run, inds_run):
            if ir is not None:
                tmp = np.zeros_like(branches_inds[bnr], dtype=bool)

                # flatten coordinates to the leaves dimension
                ir_keep = ir.astype(int).sum(axis=-1).astype(bool)
                tmp[:, :, ir_keep] = True
                # make sure leavdes that are actually not there are not counted
                tmp[~branches_inds[bnr]] = False
                inds_going_for_proposal[bnr] = tmp
            else:
                inds_going_for_proposal[bnr] = branches_inds[bnr]

            if np.any(inds_going_for_proposal[bnr]):
                at_least_one_proposal = True

            coords_going_for_proposal[bnr] = branches_coords[bnr]

        return (
            coords_going_for_proposal,
            inds_going_for_proposal,
            at_least_one_proposal,
        )

    def cleanup_proposals_gibbs(
        self,
        branch_names_run,
        inds_run,
        q,
        branches_coords,
        new_inds=None,
        branches_inds=None,
        new_branch_supps=None,
        branches_supplemental=None,
    ):
        """Set all not Gibbs-sampled parameters back

        Args:
            branch_names_run (list): List of branch names to run concurrently.
            inds_run (list): List of ``inds`` arrays including Gibbs sampling information.
            q (dict): Dictionary of new coordinate arrays for all proposal branches.
            branches_coords (dict): Dictionary of old coordinate arrays for all branches.
            new_inds (dict, optional): Dictionary of new inds arrays for all proposal branches.
            branches_inds (dict, optional): Dictionary of old inds arrays for all branches.
            new_branch_supps (dict, optional): Dictionary of new branches supplemental for all proposal branches.
            branches_supplemental (dict, optional): Dictionary of old branches supplemental for all branches.

        """
        # add back any parameters that are fixed for this round
        for bnr, ir in zip(branch_names_run, inds_run):
            if ir is not None:
                q[bnr][:, :, ~ir] = branches_coords[bnr][:, :, ~ir]

        # add other models that were not included
        for key, value in branches_coords.items():
            if key not in q:
                q[key] = value.copy()
            if new_inds is not None and key not in new_inds:
                assert branches_inds is not None
                new_inds[key] = branches_inds[key].copy()

            if new_branch_supps is not None and key not in new_branch_supps:
                assert branches_supplemental is not None
                new_branch_supps[key] = branches_supplemental[key].copy()

    def ensure_ordering(self, correct_key_order, q, new_inds, new_branch_supps):
        """Ensure proper order of key in dictionaries.

        Args:
            correct_key_order (list): Keys in correct order.
            q (dict): Dictionary of new coordinate arrays for all branches.
            new_inds (dict): Dictionary of new inds arrays for all branches.
            new_branch_supps (dict or None): Dictionary of new branches supplemental for all proposal branches.

        Returns:
            Tuple: (q, new_inds, new_branch_supps) in correct key order.

        """
        if list(q.keys()) != correct_key_order:
            q = {key: q[key] for key in correct_key_order}

        if list(new_inds.keys()) != correct_key_order:
            new_inds = {key: new_inds[key] for key in correct_key_order}

        if (
            new_branch_supps is not None
            and list(new_branch_supps.keys()) != correct_key_order
        ):
            temp = {key: None for key in correct_key_order}
            for key in new_branch_supps:
                temp[key] = new_branch_supps[key]
            new_branch_supps = deepcopy(temp)

        return q, new_inds, new_branch_supps

    def fix_logp_gibbs(self, branch_names_run, inds_run, logp, inds):
        """Set any walker with no leaves to have logp = -np.inf

        Args:
            branch_names_run (list): List of branch names to run concurrently.
            inds_run (list): List of ``inds`` arrays including Gibbs sampling information.
            logp (np.ndarray): Log of the prior going into final posterior computation.
            inds (dict): Dictionary of ``inds`` arrays for all branches.

        """
        total_leaves = np.zeros_like(logp, dtype=int)
        total_leaves_here = np.zeros_like(logp, dtype=int)
        for bnr, ir in zip(branch_names_run, inds_run):
            if ir is not None:
                tmp = np.zeros_like(inds[bnr], dtype=bool)

                # flatten coordinates to the leaves dimension
                ir_keep = ir.astype(int).sum(axis=-1).astype(bool)
                tmp[:, :, ir_keep] = True
                # make sure leaves that are actually not there are not counted
                tmp[~inds[bnr]] = False

            else:
                tmp = inds[bnr]

            total_leaves += tmp.sum(axis=-1)
            total_leaves_here += tmp.sum(axis=-1)

        for name, inds_val in inds.items():
            if name not in branch_names_run:
                total_leaves += inds_val.sum(axis=-1)

        # adjust
        logp[(total_leaves != 0) & (total_leaves_here == 0)] = -np.inf  # no use in running because no change
        logp[(total_leaves == 0) & (total_leaves_here == 0)] = 0.0  # there is nothing in the model currently

    @property
    def accepted(self):
        """Accepted counts for this move."""
        if self._accepted is None:
            raise ValueError(
                "accepted must be inititalized with the init_accepted function if you want to use it."
            )
        return self._accepted

    @accepted.setter
    def accepted(self, accepted):
        assert isinstance(accepted, np.ndarray)
        self._accepted = accepted

    @property
    def acceptance_fraction(self):
        """Acceptance fraction for this move."""
        return self.accepted / self.num_proposals

    @property
    def temperature_control(self):
        """Temperature controller"""
        return self._temperature_control

    @temperature_control.setter
    def temperature_control(self, temperature_control):
        self._temperature_control = temperature_control

        # use the setting of the temperature control to determine which log posterior function to use
        # tempered or basic
        if temperature_control is None:
            self.compute_log_posterior = self.compute_log_posterior_basic
        else:
            self.compute_log_posterior = (
                self.temperature_control.compute_log_posterior_tempered
            )

            self.ntemps = self.temperature_control.ntemps

    def compute_log_posterior_basic(self, logl, logp):
        """Compute the log of posterior

        :math:`\log{P} = \log{L} + \log{p}`

        This method is to mesh with the tempered log posterior computation.

        Args:
            logl (np.ndarray[ntemps, nwalkers]): Log-likelihood values.
            logp (np.ndarray[ntemps, nwalkers]): Log-prior values.

        Returns:
            np.ndarray[ntemps, nwalkers]: Log-Posterior values.
        """
        return logl + logp

    def tune(self, state, accepted):
        """Tune a proposal

        This is a place holder for tuning.

        Args:
            state (:class:`eryn.state.State`): Current state of sampler.
            accepted (np.ndarray[ntemps, nwalkers]): Accepted values for last pass
                through proposal.

        """
        pass

    def update(self, old_state, new_state, accepted, subset=None):
        """Update a given subset of the ensemble with an accepted proposal

        This class was updated from ``emcee`` to handle the added structure
        of Eryn.

        Args:
            old_state (:class:`eryn.state.State`): State with current information.
                New information is added to this state.
            new_state (:class:`eryn.state.State`): State with information from proposed
                points.
            accepted (np.ndarray[ntemps, nwalkers]): A vector of booleans indicating
                which walkers were accepted.
            subset (np.ndarray[ntemps, nwalkers], optional): A boolean mask
                indicating which walkers were included in the subset.
                This can be used, for example, when updating only the
                primary ensemble in a :class:`RedBlueMove`.
                (default: ``None``)

        Returns:
            :class:`eryn.state.State`: ``old_state`` with accepted points added from ``new_state``.

        """

        # TODO: update this to be use (tuples of inds) ??
        if subset is None:
            # subset of everything
            subset = np.tile(
                np.arange(old_state.log_like.shape[1]), (old_state.log_like.shape[0], 1)
            )

        # each computation is similar
        # 1. Take subset of values from old information (take_along_axis)
        # 2. Set new information
        # 3. Combine into a new temporary quantity based on accepted or not
        # 4. Put new combined subset back into full arrays (put_along_axis)

        # take_along_axis is necessary to do this all in higher dimensions
        accepted_temp = np.take_along_axis(accepted, subset, axis=1)

        # new log likelihood
        old_log_likes = np.take_along_axis(old_state.log_like, subset, axis=1)
        new_log_likes = new_state.log_like
        temp_change_log_like = new_log_likes * (accepted_temp) + old_log_likes * (
            ~accepted_temp
        )

        np.put_along_axis(old_state.log_like, subset, temp_change_log_like, axis=1)

        # new log prior
        old_log_priors = np.take_along_axis(old_state.log_prior, subset, axis=1)
        new_log_priors = new_state.log_prior.copy()

        # deal with -infs
        new_log_priors[np.isinf(new_log_priors)] = 0.0

        temp_change_log_prior = new_log_priors * (accepted_temp) + old_log_priors * (
            ~accepted_temp
        )

        np.put_along_axis(old_state.log_prior, subset, temp_change_log_prior, axis=1)

        # inds
        old_inds = {
            name: np.take_along_axis(branch.inds, subset[:, :, None], axis=1)
            for name, branch in old_state.branches.items()
        }

        new_inds = {name: branch.inds for name, branch in new_state.branches.items()}

        temp_change_inds = {
            name: new_inds[name] * (accepted_temp[:, :, None])
            + old_inds[name] * (~accepted_temp[:, :, None])
            for name in old_inds
        }

        [
            np.put_along_axis(
                old_state.branches[name].inds,
                subset[:, :, None],
                temp_change_inds[name],
                axis=1,
            )
            for name in new_inds
        ]

        # check for branches_supplemental
        run_branches_supplemental = False
        for name, value in old_state.branches_supplemental.items():
            if value is not None:
                run_branches_supplemental = True

        if run_branches_supplemental:
            # branch_supplemental
            temp_change_branch_supplemental = {}
            for name in old_state.branches:
                if old_state.branches[name].branch_supplemental is not None:
                    old_branch_supplemental = old_state.branches[
                        name
                    ].branch_supplemental.take_along_axis(
                        subset[:, :, None],
                        axis=1,
                        skip_names=self.skip_supp_names_update,
                    )
                    new_branch_supplemental = new_state.branches[
                        name
                    ].branch_supplemental[:]

                    tmp = {}
                    for key in old_branch_supplemental:
                        # need to check to see if we should skip anything
                        if key in self.skip_supp_names_update:
                            continue
                        accepted_temp_here = accepted_temp.copy()

                        # have adjust if it is an object array or a regular array
                        if new_branch_supplemental[key].dtype.name != "object":
                            for _ in range(
                                new_branch_supplemental[key].ndim
                                - accepted_temp_here.ndim
                            ):
                                accepted_temp_here = np.expand_dims(
                                    accepted_temp_here, (-1,)
                                )

                        # adjust for GPUs
                        try:
                            tmp[key] = new_branch_supplemental[key] * (
                                accepted_temp_here
                            ) + old_branch_supplemental[key] * (~accepted_temp_here)
                        except TypeError:
                            # for gpus
                            tmp[key] = new_branch_supplemental[key] * (
                                xp.asarray(accepted_temp_here)
                            ) + old_branch_supplemental[key] * (
                                xp.asarray(~accepted_temp_here)
                            )

                    temp_change_branch_supplemental[name] = BranchSupplemental(
                        tmp,
                        base_shape=new_state.branches_supplemental[name].base_shape,
                        copy=True,
                    )

                else:
                    temp_change_branch_supplemental[name] = None

            [
                old_state.branches[name].branch_supplemental.put_along_axis(
                    subset[:, :, None],
                    temp_change_branch_supplemental[name][:],
                    axis=1,
                )
                for name in new_inds
                if temp_change_branch_supplemental[name] is not None
            ]

        # sampler level supplemental
        if old_state.supplemental is not None:
            old_suppliment = old_state.supplemental.take_along_axis(subset, axis=1)
            new_suppliment = new_state.supplemental[:]

            accepted_temp_here = accepted_temp.copy()

            temp_change_suppliment = {}
            for name in old_suppliment:
                # make sure to get rid of specific supps if requested
                if name in self.skip_supp_names_update:
                    continue

                # adjust if it is not an object array
                if old_suppliment[name].dtype.name != "object":
                    for _ in range(old_suppliment[name].ndim - accepted_temp_here.ndim):
                        accepted_temp_here = np.expand_dims(accepted_temp_here, (-1,))
                try:
                    temp_change_suppliment[name] = new_suppliment[name] * (
                        accepted_temp_here
                    ) + old_suppliment[name] * (~accepted_temp_here)
                except TypeError:
                    temp_change_suppliment[name] = new_suppliment[name] * (
                        xp.asarray(accepted_temp_here)
                    ) + old_suppliment[name] * (xp.asarray(~accepted_temp_here))
            old_state.supplemental.put_along_axis(
                subset, temp_change_suppliment, axis=1
            )

        # coords
        old_coords = {
            name: np.take_along_axis(branch.coords, subset[:, :, None, None], axis=1)
            for name, branch in old_state.branches.items()
        }

        new_coords = {
            name: branch.coords for name, branch in new_state.branches.items()
        }

        # change to copy then fill due to issue of adding Nans
        temp_change_coords = {name: old_coords[name].copy() for name in old_coords}

        for name in temp_change_coords:
            temp_change_coords[name][accepted_temp] = new_coords[name][accepted_temp]

        [
            np.put_along_axis(
                old_state.branches[name].coords,
                subset[:, :, None, None],
                temp_change_coords[name],
                axis=1,
            )
            for name in new_coords
        ]

        # take care of blobs
        if new_state.blobs is not None:
            if old_state.blobs is None:
                raise ValueError(
                    "If you start sampling with a given log_like, "
                    "you also need to provide the current list of "
                    "blobs at that position."
                )

            old_blobs = np.take_along_axis(old_state.blobs, subset[:, :, None], axis=1)
            new_blobs = new_state.blobs
            temp_change_blobs = new_blobs * (accepted_temp[:, :, None]) + old_blobs * (
                ~accepted_temp[:, :, None]
            )

            np.put_along_axis(
                old_state.blobs, subset[:, :, None], temp_change_blobs, axis=1
            )

        return old_state



try:
    import cupy as xp

except (ModuleNotFoundError, ImportError) as e:
    pass

import numpy as np


class TransformContainer:
    """Container for helpful transformations

    Args:
        input_basis (list): List of integers or strings representing each 
            basis element from the input basis.
        output_basis (list): List of integers or strings representing each 
            basis element for the output basis.
        parameter_transforms (dict, optional): Keys are ``int`` or ``tuple``
            of ``int`` that contain the indexes into the parameters
            that correspond to the transformation added as the Values to the
            dict. If using ``fill_values``, you must be careful with
            making sure parameter transforms properly comes before or after
            filling values. ``int`` indicate single parameter transforms. These
            are performed first. ``tuple`` of ``int`` indicates multiple
            parameter transforms. These are performed after single-parameter transforms. 
            (default: ``None``)
        fill_dict (dict, optional): Keys must contain ``'ndim_full'``, ``'fill_inds'``,
            and ``'fill_values'``. ``'ndim_full'`` is the full last dimension of the final
            array after fill_values are added. 'fill_inds' and 'fill_values' are
            np.ndarray[number of fill values] that contain the indexes and corresponding values
            for filling. (default: ``None``)

    Raises:
        ValueError: Input information is not correct.

    """

    def __init__(self, input_basis=None, output_basis=None, parameter_transforms=None, fill_dict=None, key_map={}):

        
        # store originals
        self.original_parameter_transforms = parameter_transforms
        self.ndim_full = len(output_basis)
        self.ndim = len(input_basis)

        self.input_basis, self.output_basis = input_basis, output_basis

        test_inds = []
        for key in input_basis:
            if key not in output_basis and key not in key_map:
                raise ValueError("All keys in input_basis must be present in output basis, or you must provide a key_map")
            key_in = key if key not in key_map else key_map[key]
            test_inds.append(output_basis.index(key_in))

        self.test_inds = test_inds = np.asarray(test_inds)
        if parameter_transforms is not None:
            # differentiate between single and multi parameter transformations
            self.base_transforms = {"single_param": {}, "mult_param": {}}

            # iterate through transforms and setup single and multiparameter transforms
            for key, item in parameter_transforms.items():
                if isinstance(key, str) or isinstance(key, int):
                    if key not in output_basis:
                        assert key in key_map
                        key = key_map[key]
                    key_in = output_basis.index(key)
                    self.base_transforms["single_param"][key_in] = item
                elif isinstance(key, tuple):
                    _tmp = []
                    for i in range(len(key)):
                        key_tmp = key[i]
                        if  key_tmp not in output_basis:
                            assert key_tmp in key_map
                            key_tmp = key_map[key_tmp]
                        _tmp.append(output_basis.index(key_tmp))
                    self.base_transforms["mult_param"][tuple(_tmp)] = item
                else:
                    raise ValueError(
                        "Parameter transform keys must be str (or int) or tuple of strs (or ints). {} is neither.".format(
                            key
                        )
                    )
        else:
            self.base_transforms = None

        self.original_fill_dict = fill_dict
        if fill_dict is not None:
            if not isinstance(fill_dict, dict):
                raise ValueError("fill_dict must be a dictionary.")

            self.fill_dict = {}
            self.fill_dict["fill_inds"] = []
            self.fill_dict["fill_values"] = []
            for key in fill_dict.keys():
                self.fill_dict["fill_inds"].append(output_basis.index(key))
                self.fill_dict["fill_values"].append(fill_dict[key])

            # set up test_inds accordingly
            self.fill_dict["test_inds"] = test_inds
            self.fill_dict["fill_inds"] = np.asarray(self.fill_dict["fill_inds"])
            self.fill_dict["fill_values"] = np.asarray(self.fill_dict["fill_values"])

        else:
            self.fill_dict = None

    def transform_base_parameters(
        self, params, copy=True, return_transpose=False, xp=None
    ):
        """Transform the base parameters

        Args:
            params (np.ndarray[..., ndim]): Array with coordinates. This array is
                transformed according to the ``self.base_transforms`` dictionary.
            copy (bool, optional): If True, copy the input array.
                (default: ``True``)
            return_transpose (bool, optional): If True, return the transpose of the
                array. (default: ``False``)
            xp (object, optional): ``numpy`` or ``cupy``. If ``None``, use ``numpy``.
                (default: ``None``) 

        Returns:
            np.ndarray[..., ndim]: Transformed ``params`` array.

        """

        # cupy or numpy
        if xp is None:
            xp = np

        if self.base_transforms is not None:
            params_temp = params.copy() if copy else params
            params_temp = params_temp.T
            # regular transforms
            for ind, trans_fn in self.base_transforms["single_param"].items():
                params_temp[ind] = trans_fn(params_temp[ind])

            # multi parameter transforms
            for inds, trans_fn in self.base_transforms["mult_param"].items():
                temp = trans_fn(*[params_temp[i] for i in inds])
                for j, i in enumerate(inds):
                    params_temp[i] = temp[j]

            # its actually the opposite now
            if return_transpose:
                return params_temp
            else:
                return params_temp.T

        else:
            if return_transpose:
                return params.T
            else:
                return params

    def fill_values(self, params, xp=None):
        """fill fixed parameters

        This also adjusts parameter order as needed between the two bases. 

        Args:
            params (np.ndarray[..., ndim]): Array with coordinates. This array is
                filled with values according to the ``self.fill_dict`` dictionary.
            xp (object, optional): ``numpy`` or ``cupy``. If ``None``, use ``numpy``.
                (default: ``None``) 

        Returns:
            np.ndarray[..., ndim_full]: Filled ``params`` array.

        """
        if self.fill_dict is not None:
            if xp is None:
                xp = np

            # get shape
            shape = params.shape

            # setup new array to fill
            params_filled = xp.zeros(shape[:-1] + (self.ndim_full,))
            test_inds = xp.asarray(self.fill_dict["test_inds"])
            # special indexing to properly fill array with params
            indexing_test_inds = tuple([slice(0, temp) for temp in shape[:-1]]) + (
                test_inds,
            )

            # fill values directly from params array
            params_filled[indexing_test_inds] = params

            fill_inds = xp.asarray(self.fill_dict["fill_inds"])
            # special indexing to fill fill_values
            indexing_fill_inds = tuple([slice(0, temp) for temp in shape[:-1]]) + (
                fill_inds,
            )

            # add fill_values at fill_inds
            params_filled[indexing_fill_inds] = xp.asarray(
                self.fill_dict["fill_values"]
            )

            return params_filled

        else:
            return params

    def both_transforms(
        self, params, copy=True, return_transpose=False, xp=None
    ):
        """Transform the parameters and fill fixed parameters

        This fills the fixed parameters and then transforms all of them. Therefore, the user
        must be careful with the indexes input. 

        This is generally the direction recommended because fixed parameters may change
        non-fixed parameters during parameter transformations. This can be reversed
        with the ``reverse`` kwarg.

        Args:
            params (np.ndarray[..., ndim]): Array with coordinates. This array is
                transformed according to the ``self.base_transforms`` dictionary.
            copy (bool, optional): If True, copy the input array.
                (default: ``True``)
            return_transpose (bool, optional): If ``True``, return the transpose of the
                array. (default: ``False``)
            xp (object, optional): ``numpy`` or ``cupy``. If ``None``, use ``numpy``.
                (default: ``None``) 

        Returns:
            np.ndarray[..., ndim]: Transformed and filleds ``params`` array.

        """
        # numpy or cupy
        if xp is None:
            xp = np

        # run transforms first
        temp = self.fill_values(params, xp=xp)
        temp = self.transform_base_parameters(
            temp, copy=copy, return_transpose=return_transpose, xp=xp
        )
        return temp
